#include <ctype.h>
#include <string>
#include <iostream>
#include <fstream>
#include <strstream>
#include "gate.H"
#include "gen.H"
#include "LED.H"
#include "connector.H"

// Simple stream-reader object which returns the next string.  It
// mostly skips /* comments */  Note that comment must be surrounded 
// by white space (unlike a real language).  The reader doesn't buffer,
// so we can switch back and forth with the base stream.
class reader {
private:
	// Base input stream.
	istream &in;

	// Inside a comment.
	bool incom;
public:
	reader(istream &_in): in(_in), incom(false) { }
	operator bool() { return in.good(); }
	string next() {
		string ret = "";
		while(in >> ret) {
			if(incom) {
				if(ret.length() >= 2 &&
				   ret.substr(ret.length() - 2, 2) == "*/")
					incom = false;
				continue;
			} else {
				if(ret.substr(0,2) == "/*") {
					int len = ret.length();
					if(len >= 2 && 
					   ret.substr(len - 2, 2) != "*/")
						incom = true;
					continue;
				}
			}
			break;
		}

		return ret;
	}
};

/* Let reader work with >> for the types we care about.  Note also that
   reading a double behaves differently from real streams when the
   integer is not followed by a space. */
reader & operator >> (reader &r, string &s)
{
	s = r.next();
	return r;
}

reader & operator >> (reader &r, int &i)
{
	string s;
	r >> s;
	istrstream(s.c_str()) >> i;
	return r;
}

reader & operator >> (reader &r, double &d)
{
	string s;
	r >> s;
	istrstream(s.c_str()) >> d;
	return r;
}

//
// Input language is tokenized by space separation, so a lot of things that
// should work, don't.  Add spaces.
// /* comments */
// dev1 -> dev2@p
//	Link from dev1 to dev2, device specs are inst1.inst2....Xid, where
//	X is a device number.  Port is letter a-z, defaults to a.
// [[ dev opts ]]
// 	Set the option string for device Xm.  The option words are built
//	back into a string speparated by spaces and delived to the simulation
// 	object, which processes at it wishes.  Presently, the clock, 
//	generator, and LED pay attention to options.  The clock treats the
//	string as two integers, the period and the high time.  The generator
//	treats the string as the min and max period between changes.  The
//	LED treats it as a label printed instead of the "LED m" when the
//	it changes its display.
// name = { whatever }
//	Creates a subcircuit with a given name.  
// iname : name n
//	Create n copies of circuit named name.  Instances iname[0]..iname[n-1].
//	Reference into an instance are qualified as name.X.id
//

// List of legal letters.
const string Connector::gatelets = "AOXNGCL";

Connector::Connector(char *fn)
{
	ifstream ifs(fn);

	if(!ifs) {
		cerr << "Cannot read " << fn << endl;
		exit(9);
	}

	// Simulation stop time.
	reader in(ifs);
	in >> stoptime;
	if(stoptime <= 0.0) {
		cerr << "Stop time is missing or non-positive." << endl;
		exit(3);
	}

	parent = 0;
	build(ifs);

	// Symbol table only needed durning read.
	clear_symbols();
}

/* Map a gate reference to a device id. */
int Connector::mapgate(string name, char &type, int &num, bool maycreate)
{
	/* Look for a sub-instance. */
	string::size_type loc = name.find(".");
	if(loc == string::npos) {
		// Look up in this.
		type = name[0];
		maxes_t::iterator mi = maxes.find(type);
		if(mi == maxes.end()) return 0;
		gatemap_t::iterator gi = gatemap.find(name);
		if(gi == gatemap.end()) {
			if(maycreate) {
				// Add a new one.
				num = ++mi->second;
				gatemap[name] = num;
				return 1;
			} else 
				return 0;
		} else {
			// Ret the old one.
			num = gi->second;
			return 1;
		}
	} else {
		// Look at the instance name.
		string iname = name.substr(0,loc);
		name = name.substr(loc + 1);

		// Find the instance
		instmap_t::iterator inst = instmap.find(iname);
		if(inst == instmap.end()) return 0;

		// Recursively look up the name.
		if(!inst->second.c->mapgate(name, type, num)) return 0;

		// Adjust for starting offset of the instance.
		num += inst->second.starts[gatelets.find(type)];
		return 1;
	}
}


void Connector::build(istream &ins)
{
	reader in(ins);

	for(int i = gatelets.length(); i--; )
		maxes[gatelets[i]] = -1;
	//maxes['A'] = -1;
	//maxes['O'] = -1;
	//maxes['X'] = -1;
	//maxes['N'] = -1;
	//maxes['G'] = -1;
	//maxes['C'] = -1;
	//maxes['L'] = -1;

	string from, arrow, dest;
	while(in >> from) {
		// Termination.
		if(parent && from == "}") return;
		if(!(in >> arrow >> dest)) {
			cerr << "Missing token after " << from << endl;
			exit(41);
		}

		// Build a def.
		if(arrow == "=" && dest == "{") {
			symtab[from] = new Connector(ins, this);
			continue;
		}

		// Find a use of a named subcircuit.
		// iname : name n
		if(arrow == ":") {
			int ninst;
			in >> ninst;

			symtab_t::iterator i = symtab.find(dest);
			if(i == symtab.end()) {
				Connector *p = parent;
				for(; p != 0; p = p->parent) {
					i = p->symtab.find(dest);
					if(i != p->symtab.end()) break;
				}

				if(p == 0) {
					cerr << "Use of undefined subcircuit "
					     << dest << endl;
					exit(45);
				}
			}

			for(int n = 0; n < ninst; n++) {
				ostrstream nbuf;
				nbuf << from << "[" << n << "]" << ends;
				insert(i->second, n, nbuf.str());
			}
			
			continue;
		}

		// Check for a parm setting.
		if(parent == 0 && from == "[[") {
			char ltyp;
			int serial;
			if(!mapgate(arrow, ltyp, serial, true)) {
				cerr << "Bad gate id in " << dest << endl;
				exit(17);
			}

			// Build the parm string.
			string parms = "";
			if(dest != "]]") {
				parms = dest;
				string p;
				while(in >> p) {
					if(p == "]]") break;
					parms += " " + p;
				}
			}

			parmlist[dev_id_t(ltyp, serial)] = parms;

			continue;
		}

		// Connection
		if(arrow != "->") {
			cerr << "No -> in \"" << from << " " << 
				arrow << " " << dest << "\"" << endl;
			exit(10);
		}

		// Source type and number.
		char frtype;
		int fromno;
		if(!mapgate(from, frtype, fromno, true)) {
			cerr << "Bad gate id in " << from << endl;
			exit(11);
		}

		// Dest type, number and port.
		char port = 'a';
		int dlen = dest.length();
		if(dlen > 3 && dest.substr(dlen-2, 1) == "@") {
			port = dest.substr(dlen-1)[0];
			if(!isascii(port) || !islower(port)) {
				cerr << "Bad port number in \"" << 
					dest << "\"" << endl;
				exit(4);
			}
			dest = dest.substr(0,dlen-2);
		}
		char totype;
		int tono;
		if(!mapgate(dest, totype, tono, true)) {
			cerr << "Bad gate id in " << dest << endl;
			exit(12);
		}

		// Okay.  Add it.
		addcon(frtype, fromno, totype, tono, port);
	}
}

// Add a connection to the object
void Connector::addcon(char ftype, int fromno, char ttype, int tono, char port)
{
	// Check this beast.
	if(fromno < 0) {
		cerr << "Attempt to insert negative source serial (" << fromno
		     << ")." << endl;
		exit(5);
	}
	if(tono < 0) {
		cerr << "Attempt to insert negative source serial." << tono
		     << ")." << endl;
		exit(6);
	}

	// Update the device counts.
	if(maxes[ttype] < tono) maxes[ttype] = tono;
	if(maxes[ftype] < fromno) maxes[ftype] = fromno;

	// Update the destination input size.
	int dinsize = port - 'a' + 1;
	dev_id_t dst(ttype, tono);
	insizes_t::iterator si = insizes.find(dst);
	if(si == insizes.end()) 
		insizes.insert(insizes_t::value_type(dst, dinsize));
	else if(si->second < dinsize) 
		si->second = dinsize;

	// Record the connection.
	dest_t drec = { ttype, tono, port };
	dev_id_t src(ftype, fromno);
	lookup_map_t::iterator lp = lookup_map.find(src);
	if(lp == lookup_map.end()) {
		lookup_t *vp = new lookup_t;
		vp->push_back(drec);
		lookup_map[src] = vp;
	} else
		lp->second->push_back(drec);
}

void Connector::xmit(char srctype, int serial, double when, int signal)
{
	if(when > stoptime) return;

	dev_id_t src(srctype, serial);
	lookup_map_t::iterator list = lookup_map.find(src);
	if(list == lookup_map.end()) return;
	lookup_t::iterator j = list->second->begin(); 
	for(;j != list->second->end(); ++j) {
		dest_t dst = *j;
		switch(dst.dtype) {
		case 'A':
			SCHEDULE_a_signal
				(when, SpGetObjHandle("S_And_Gate_MGR", 
						      dst.serial),
				 dst.port, signal);
			break;
		case 'O':
			SCHEDULE_o_signal
				(when, SpGetObjHandle("S_Or_Gate_MGR", 
						      dst.serial),
				 dst.port, signal);
			break;
		case 'X':
			SCHEDULE_x_signal
				(when, SpGetObjHandle("S_Xor_Gate_MGR", 
						      dst.serial),
				 dst.port, signal);
			break;
		case 'N':
			SCHEDULE_n_signal
				(when, SpGetObjHandle("S_Not_Gate_MGR", 
						      dst.serial),
				 signal);
			break;
		case 'G':
			break;
		case 'C':
			break;
		case 'L':
			SCHEDULE_l_signal
				(when, SpGetObjHandle("S_LED_MGR", 
						      dst.serial),
				 signal);
			break;
		}
	}

}

string Connector::parms(char type, int serial)
{
	parmlist_t::iterator i = parmlist.find(dev_id_t(type, serial));
	if(i == parmlist.end()) return "";
	return i->second;
}

/* Insert a subcircuit. */
void Connector::insert(Connector *c, int n, string iname)
{
	// Create an instance record.
	instmap.insert(instmap_t::value_type(iname, subinst_t(c, this)));
	instmap_t::iterator ii = instmap.find(iname);
	
	// Add the links.
	lookup_map_t::iterator i = c->lookup_map.begin();
	for(;i != c->lookup_map.end(); ++i) {
		char srctype = i->first.first;
		int from = i->first.second;
		lookup_t::iterator j = i->second->begin();
		for(;j != i->second->end(); ++j)
			addcon(srctype, 
			       ii->second.starts[gatelets.find(srctype)] + 
			       		from,
			       j->dtype, 
			       ii->second.starts[gatelets.find(j->dtype)] +
			       		j->serial,
			       j->port);
	}
}

/* Cleanup. */
void Connector::clear_symbols() 
{
	symtab_t::iterator i = symtab.begin();
	for(;i != symtab.end(); ++i) {
		delete i->second;
	}
	symtab.clear();

	gatemap.clear();
}
