#ifndef CONNECTOR_H
#define CONNECTOR_H

#include <map>
#include <vector>
#include <string>

using namespace std;

/* Number of device letters. */
#define NUMLETS 7

class Connector {
public:
	// Construct the object with the contents of file.  Exits 
	// on failure.
	Connector(char *fn = "conn.lines");

	// Transmit from a given source type (AOXNGL) and serial number to
	// to all devices it is connected to according to the file.
	// The signal is scheduled at the indicated time for all connected
	// devices, so long as when is not larger than the stop time.
	void xmit(char srctype, int serial, double when, int signal);

	// Get the number of devices of the given type (AOXNGL).
	int count(char type) { return maxes[type] + 1; }

	// Get the input size of the device of the given type (AOXNGL)
	// and serial number.
	int insize(char type, int serial) { 
		return insizes[dev_id_t(type, serial)];
	}

	// Get the stop time.
	double stopat() { return stoptime; }

	// Get parameter data from the input file, if any, or the empty
	// string.
	string parms(char type, int serial);

	// Cleanup.
	void clear_symbols();
	~Connector() { clear_symbols(); }

	// List of gate letters and the number of them.
	static const string gatelets; // = "AOXNGCL";
private:
	// Parent pointer used for def lookup.
	Connector *parent;

	// Most of the work for the constructor.  The internal flag is for
	// recusion to read subcircuits.
	void build(istream &);

	// This is used to construct subcircuits.
	Connector(istream &strm, Connector *p) { 
		parent = p;
		build(strm); 
	}

	// Add a connection to the object
	void Connector::addcon(char ftype, int fromno, 
			       char ttype, int tono, char port);

	// Stop time from the file.
	double stoptime;

	// Network of connections.
	//   lookup_map[type, serial] = vector of <type, serial, port>
	// The type and serial values are joined in an STL pair object
	// so they can be the key of map.  The type serial and port are
	// combined 
	struct dest_t { char dtype; int serial; char port; };
	typedef vector<dest_t> lookup_t;
	typedef pair<char, int> dev_id_t;
	typedef map<dev_id_t, lookup_t *> lookup_map_t;
	lookup_map_t lookup_map;

	// Max number of devices of each type.
	typedef map<char, int> maxes_t;
	maxes_t maxes;

	// Input size of each device.
	typedef map<dev_id_t, int> insizes_t;
	insizes_t insizes;

	// Parameter messages.
	typedef map<dev_id_t, string> parmlist_t;
	parmlist_t parmlist;

	// Dictionary of defined subcircuits.  
	typedef map<string, Connector *> symtab_t;
	symtab_t symtab;

	// Dictionary of subcircuit instances.  start are 
	// A, O, X, N, G, C, L; 
	struct subinst_t { 
		Connector *c;
		int starts[NUMLETS]; 
		subinst_t(Connector *rmt, Connector *me) { 
			c = rmt;
			for(int i = 0; i < NUMLETS; i++)
				starts[i] = me->count(me->gatelets[i]);
		}
	};
	typedef map<string, subinst_t> instmap_t;
	instmap_t instmap;

	// Gate name table.  Name is type and name together.
	typedef map<string, int> gatemap_t;
	gatemap_t gatemap;

	// Map a gate reference.
	int mapgate(string name, char &, int &, bool maycreate = false);

	// Insert a named subcircuit.
	void insert(Connector *c, int n, string name);
};

extern Connector *conobj;

#endif
